(********************************************************************)
(*                                                                  *)
(*  The LustreC compiler toolset   /  The LustreC Development Team  *)
(*  Copyright 2012 -    --   ONERA - CNRS - INPT                    *)
(*                                                                  *)
(*  LustreC is free software, distributed WITHOUT ANY WARRANTY      *)
(*  under the terms of the GNU Lesser General Public License        *)
(*  version 2.1.                                                    *)
(*                                                                  *)
(********************************************************************)

open Utils
open Lustre_types
open Corelang
open Causality
open Scheduling_type

(* Topological sort with a priority for variables belonging in the same equation
   lhs. For variables still unrelated, standard compare is used to choose the
   minimal element. This priority is used since it helps a lot in factorizing
   generated code. Moreover, the dependency graph is browsed in a depth-first
   manner whenever possible, to improve the behavior of optimization algorithms
   applied in forthcoming compilation steps. In the following functions: -
   [eq_equiv] is the equivalence relation between vars of the same equation lhs
   - [g] the (imperative) graph to be topologically sorted - [pending] is the
   set of unsorted root variables so far, equivalent to the last sorted var -
   [frontier] is the set of unsorted root variables so far, not belonging in
   [pending] - [sort] is the resulting topological order.
   Also, the chosen element at each iteration is the smallest one, wrt the given
   ordering. This ordering is given by the functor SchedFanOrderMake 
   wrt the dependency graph.
*)

module type SchedStrat =
  sig
    type t
    type set

    val empty : IdentDepGraph.t -> set
    val is_empty : set -> bool
    val add_init : ident -> set -> set
    val add_succ : t -> ident -> set -> set
    val remove : t -> set -> set
    val choose : set -> t
    val partition : (t -> bool) -> set -> set * set
    val vertex : t -> ident
  end

module Naive : SchedStrat =
  struct
    type t = ident
    type set = ISet.t

    let empty _ = ISet.empty
    let is_empty set = ISet.is_empty set
    let add_init x set = ISet.add x set
    let add_succ _ tgt set = ISet.add tgt set
    let remove x set = ISet.remove x set
    let choose set = ISet.min_elt set
    let partition p set = ISet.partition p set
    let vertex x = x
  end

module Fanout : SchedStrat =
  struct
    let compare (fan1, x1) (fan2, x2) =
      match fan1 - fan2 with
        0 -> Stdlib.compare x1 x2
      | c -> c
    
    module ISet = Set.Make (struct type t = int * ident let compare = compare end)

    type t = int * ident
    type set = IdentDepGraph.t * ISet.t

    let empty g = (g, ISet.empty)
    let is_empty (_, set) = ISet.is_empty set
    let add_init x (g, set) =
      let fan = IdentDepGraph.out_degree g x in
      (g, ISet.add (fan, x) set)
    let add_succ _ tgt (g, set) =
      let fan = IdentDepGraph.out_degree g tgt in
      (g, ISet.add (fan, tgt) set)
    let remove (fan, x) (g, set) =
      begin
        assert (ISet.mem (fan, x) set);
        (g, ISet.remove (fan, x) set)
      end
    let choose (_, set) = ISet.min_elt set
    let partition p (g, set) =
      let a, b = ISet.partition p set
      in (g, a), (g, b)
    let vertex (_, x) = x
  end

module Fanin : SchedStrat =
  struct
    let compare (fan1, x1) (fan2, x2) =
      match fan1 - fan2 with
        0 -> Stdlib.compare x1 x2
      | c -> c
    
    module ISet = Set.Make (struct type t = int * ident let compare = compare end)

    type t = int * ident
    type set = int IMap.t * ISet.t

    let empty g =
      let fanins = IdentDepGraph.fold_vertex (fun v -> IMap.add v (IdentDepGraph.in_degree g v)) g IMap.empty
      in (fanins, ISet.empty)
    let is_empty (_, set) = ISet.is_empty set
    let add_init x (map, set) =
      (map, ISet.add (0, x) set)
    let add_succ _ tgt (map, set) =
      let fan = IMap.find tgt map in
      (map, ISet.add (fan, tgt) set)
    let remove (fan, x) (g, set) =
      begin
        assert (ISet.mem (fan, x) set);
        (g, ISet.remove (fan, x) set)
      end
    let choose (_, set) = ISet.min_elt set
    let partition p (map, set) =
      let a, b = ISet.partition p set
      in (map, a), (map, b)
    let vertex (_, x) = x
  end

module Depth : SchedStrat =
  struct
    module DMap = Map.Make (struct type t = int let compare x1 x2 = x1 - x2 end)

    type t = ident
    type set = t list

    let empty _ = []
    let is_empty set = set = []
    let add x set =
      x::(List.filter (fun y -> not (y = x)) set)
    let add_init x set = x::set
    let add_succ _ tgt set = add tgt set
    let remove x set =
      begin
        assert (List.mem x set);
        List.filter (fun y -> not (y = x)) set
      end
    let choose set = List.hd set
    let partition p set = List.partition p set
    let vertex x = x
  end
   
let register_strategy, find_strategy =
  let strategies = ref ([] : (string * (module SchedStrat)) list) in
  (fun name strat -> strategies := (name, strat) :: !strategies),
  (fun name       -> try List.assoc name !strategies
                     with Not_found -> (module Naive : SchedStrat))

let _ =
  begin
    register_strategy "naive"  (module Naive  : SchedStrat);
    register_strategy "fanout" (module Fanout : SchedStrat);
    register_strategy "fanin"  (module Fanin  : SchedStrat);
    register_strategy "depth"  (module Depth  : SchedStrat);
  end

module Scheduler =
  struct
(* Checks whether the currently scheduled variable [choice] is an output of a
   call, possibly among others *)
let is_call_output choice g =
  List.exists ExprDep.is_call_instance_var (IdentDepGraph.succ g choice)

let topological_sort =
  let module ISet = (val find_strategy !Options.scheduling) in
  (* Adds successors of [v] in graph [g] in [pending] or [frontier] sets, wrt
     [eq_equiv], then removes [v] from [g] *)
  let add_successors eq_equiv g src pending frontier =
    let v = ISet.vertex src in
    let succs_v = IdentDepGraph.succ g v in
    IdentDepGraph.remove_vertex g v;
    List.iter
      (fun v' ->
        if is_graph_root v' g then
          if eq_equiv v v' then pending := ISet.add_succ src v' !pending
          else frontier := ISet.add_succ src v' !frontier)
      succs_v
  in
  (* Chooses the next var to be sorted, taking priority into account. Modifies
     [pending] and [frontier] accordingly. *)
  let next_element eq_equiv g sort call pending frontier =
    if ISet.is_empty !pending then (
      let choice = ISet.choose !frontier in
      let vertex = ISet.vertex choice in
      (*Format.eprintf "-1-> %s@." choice;*)
      frontier := ISet.remove choice !frontier;
      let p, f = ISet.partition (fun elt -> eq_equiv vertex (ISet.vertex elt)) !frontier in
      pending := p;
      frontier := f;
      call := is_call_output vertex g;
      (*if !call then Format.eprintf "call var@."*)
      add_successors eq_equiv g choice pending frontier;
      if not (ExprDep.is_ghost_var vertex) then sort := [ vertex ] :: !sort)
    else
      let choice = ISet.choose !pending in
      let vertex = ISet.vertex choice in
      (*Format.eprintf "-2-> %s@." choice;*)
      pending := ISet.remove choice !pending;
      add_successors eq_equiv g choice pending frontier;
      (*if !call then Format.eprintf "call var@."*)
      if not (ExprDep.is_ghost_var vertex) then
        sort :=
          if !call then (vertex :: List.hd !sort) :: List.tl !sort
          else [ vertex ] :: !sort
  in
  (* Topological sort of dependency graph [g], with priority. *)
  let topological_sort eq_equiv g =
    let roots = graph_roots g in
    assert (roots <> []);
    let call = ref false in
    let frontier = ref (List.fold_right ISet.add_init roots (ISet.empty g)) in
    let pending = ref (ISet.empty g) in
    let sorted = ref [] in
    while not (ISet.is_empty !frontier && ISet.is_empty !pending) do
      (*Format.eprintf "frontier = {%a}, pending = {%a}@." (fun fmt -> ISet.iter
        (fun e -> Format.pp_print_string fmt e)) !frontier (fun fmt -> ISet.iter
        (fun e -> Format.pp_print_string fmt e)) !pending;*)
      next_element eq_equiv g sorted call pending frontier
    done;
    IdentDepGraph.clear g;
    (*List.iter (fun l -> Format.eprintf "(%a) " (fun fmt -> List.iter (fun v -> Format.pp_print_string fmt v)) l) !sorted;*)
    !sorted
  in topological_sort

  end

(* XXX: UNUSED *)
(* Filters out normalization variables and renames instance variables to keep
   things readable, in a case of a dependency error *)
(* let filter_original n vl =
 *   List.fold_right
 *     (fun v res ->
 *       if ExprDep.is_instance_var v then
 *         Format.sprintf "node %s" (ExprDep.undo_instance_var v) :: res
 *       else
 *         let vdecl = get_node_var v n in
 *         if vdecl.var_orig then v :: res else res)
 *     vl [] *)

let eq_equiv eq_equiv_hash v1 v2 =
  try Hashtbl.find eq_equiv_hash v1 = Hashtbl.find eq_equiv_hash v2
  with Not_found -> false

let schedule_node n =
  (* let node_vars = get_node_vars n in *)
  Log.report ~level:5 (fun fmt ->
      Format.fprintf fmt "scheduling node %s@ " n.node_id);
  let eq_equiv = eq_equiv (ExprDep.node_eq_equiv n) in

  let node, g = global_dependency n in

  (* TODO X: extend the graph with inputs (adapt the causality analysis to deal
     with inputs compute: coi predecessors of outputs warning (no modification)
     when memories are non used (do not impact output) or when inputs are not
     used (do not impact output) DONE ! *)
  let dep_graph = IdentDepGraph.copy g in
  let schedule = Scheduler.topological_sort eq_equiv g in
  let unused_vars = Liveness.compute_unused_variables n dep_graph in
  let fanin_table = Liveness.compute_fanin n dep_graph in
  { node; schedule; unused_vars; fanin_table; dep_graph }

(* let schedule_eqs eqs =
 *   let eq_equiv = eq_equiv (ExprDep.eqs_eq_equiv eqs) in
 *   assert false (\* TODO: continue to implement scheduling of eqs for spec *\) *)

let compute_node_reuse_table report =
  let disjoint = Disjunction.clock_disjoint_map (get_node_vars report.node) in
  let reuse =
    Liveness.compute_reuse_policy
      report.node
      report.schedule
      disjoint
      report.dep_graph
  in
  (* if !Options.print_reuse then begin Log.report ~level:0 (fun fmt ->
     Format.fprintf fmt "OPT:%B@." (try (Hashtbl.iter (fun s1 v2 -> if s1 =
     v2.var_id then raise Not_found) reuse; false) with Not_found -> true) );
     Log.report ~level:0 (fun fmt -> Format.fprintf fmt "OPT:clock disjoint map
     for node %s: %a" n'.node_id Disjunction.pp_disjoint_map disjoint );
     Log.report ~level:0 (fun fmt -> Format.fprintf fmt "OPT:reuse policy for
     node %s: %a" n'.node_id Liveness.pp_reuse_policy reuse ); end; *)
  Hashtbl.filter_map_inplace
    (fun x v -> if x = v.var_id then None else Some v)
    reuse;
  reuse

let schedule_prog prog =
  List.fold_right
    (fun top_decl (accu_prog, sch_map) ->
      match top_decl.top_decl_desc with
      | Node nd ->
        let report = schedule_node nd in
        ( { top_decl with top_decl_desc = Node report.node } :: accu_prog,
          IMap.add nd.node_id report sch_map )
      | _ ->
        top_decl :: accu_prog, sch_map)
    prog
    ([], IMap.empty)

let compute_prog_reuse_table report = IMap.map compute_node_reuse_table report

(* removes inlined local variables from schedule report, which are now
   useless *)
let remove_node_inlined_consts_and_locals consts consts_and_locals report =
  let is_inlined v = IMap.mem v consts_and_locals in
  let schedule' =
    List.fold_right
      (fun heads q ->
        let heads' = List.filter (fun v -> not (is_inlined v)) heads in
        if heads' = [] then q else heads' :: q)
      report.schedule
      []
  in
  IMap.iter (fun v _ -> Hashtbl.remove report.fanin_table v) consts_and_locals;
  IMap.iter
    (fun v _ ->
      if List.exists (fun decl -> Corelang.(const_of_top decl).const_id = v)
           consts
      then
        let iv = ExprDep.(mk_read_var v) in
        IdentDepGraph.remove_vertex report.dep_graph iv
      else
        let iv = ExprDep.(mk_call_instance_var (mk_call_instance_var v)) in
        Liveness.replace_in_dep_graph v iv report.dep_graph)
    consts_and_locals;
  { report with schedule = schedule' }

let remove_prog_inlined_consts_and_locals consts removed reuse =
  IMap.mapi (fun id -> remove_node_inlined_consts_and_locals consts (IMap.find id removed)) reuse

let pp_eq_schedule fmt vl =
  match vl with
  | [] ->
    assert false
  | [ v ] ->
    Format.pp_print_string fmt v
  | _ ->
    Format.fprintf fmt "(%a)" (Format.pp_comma_list Format.pp_print_string) vl

let pp_schedule fmt node_schs =
  IMap.iter
    (fun nd report ->
      Format.(
        fprintf
          fmt
          "%s schedule: %a@ "
          nd
          (pp_print_list ~pp_sep:pp_print_semicolon pp_eq_schedule)
          report.schedule))
    node_schs

let pp_fanin_table fmt node_schs =
  IMap.iter
    (fun nd report ->
      Format.fprintf fmt "%s: %a@ " nd Liveness.pp_fanin report.fanin_table)
    node_schs

let pp_dep_graph fmt node_schs =
  IMap.iter
    (fun nd report ->
      Format.fprintf
        fmt
        "%s dependency graph: %a@ "
        nd
        pp_dep_graph
        report.dep_graph)
    node_schs

let pp_warning_unused fmt node_schs =
  IMap.iter
    (fun nd report ->
      let unused = report.unused_vars in
      if not (ISet.is_empty unused) then
        let nd =
          match (Corelang.node_from_name nd).top_decl_desc with
          | Node nd ->
            nd
          | _ ->
            assert false
        in
        ISet.iter
          (fun u ->
            let vu = get_node_var u nd in
            if vu.var_orig then
              Error.pp_warning vu.var_loc
                (fun fmt ->
                  Format.fprintf
                    fmt
                    "variable '%s' seems unused" u))
          unused)
    node_schs

(* Sort eqs according to schedule *)
(* Sort the set of equations of node [nd] according to the computed schedule
   [sch] *)
let sort_equations_from_schedule eqs sch =
  Log.report ~level:10 (fun fmt ->
      Format.fprintf
        fmt
        "schedule: %a@ "
        (Format.pp_print_list ~pp_sep:Format.pp_print_semicolon pp_eq_schedule)
        sch);
  let split_eqs = Splitting.tuple_split_eq_list eqs in
  (* Flatten schedule *)
  let sch =
    List.fold_right (fun vl res -> List.map (fun v -> [ v ]) vl @ res) sch []
  in
  let eqs_rev, remainder =
    List.fold_left
      (fun (accu, node_eqs_remainder) vl ->
        (* For each variable in vl, there should exists the equations in accu *)
        if
          List.for_all
            (fun v -> List.exists (fun eq -> List.mem v eq.eq_lhs) accu)
            vl
        then accu, node_eqs_remainder
        else
          let eq_v, remainder = find_eq vl node_eqs_remainder in
          eq_v :: accu, remainder)
      ([], split_eqs)
      sch
  in
  let eqs = List.rev eqs_rev in
  let unused =
    if List.length remainder > 0 then (
      Log.report ~level:3 (fun fmt ->
          Format.fprintf
            fmt
            "[Warning] Equations not used are@ %a@ Full equation set is:@ %a@ "
            Printers.pp_node_eqs
            remainder
            Printers.pp_node_eqs
            eqs);
      let vars =
        List.fold_left (fun accu eq -> eq.eq_lhs @ accu) [] remainder
      in
      Log.report ~level:1 (fun fmt ->
          Format.fprintf
            fmt
            "[Warning] Unused variables: %a@ "
            (Format.pp_comma_list Format.pp_print_string)
            vars);
      vars)
    else []
  in
  eqs, unused

(* Local Variables: *)
(* compile-command:"make -C .." *)
(* End: *)
